(* Imperative Program Simulator *)
(* By Siddharth Eswaran | 2005CS10187 *)
(* CSL302 Programming Languages *)

signature SIMULATOR =
sig
	
	datatype frame = NULL | F of {name:string, vfp:(frame ref),caller : string, vars:((string * int ref) list)}
	datatype 'a tree = Node of 'a * ('a tree list)

	val buildTree : DataTypes.procedure -> string tree 
	val buildFrame : DataTypes.procedure -> frame ref -> string -> frame
	(* Loads and initialises Simulator *)
	val load : DataTypes.procedure -> unit 

	(* The main program. Simulates calling execution and return
	* of procedures using stack-based-allocation model.
	*)  
	val start : unit -> unit

	(* Resets all internal variables. *)
	val reset : unit -> unit
end;

structure ImpSim:SIMULATOR =
struct
open DataTypes;
open TextIO;
open Exp;
open BuildAst;

	(* DATATYPE frame : Represents the frame of a procedure on the stack.
	* Contains the loacl variables, parameters (by POC) and a static link to
	* the parent procedure.
	*) 
	datatype frame = NULL | F of {name:string, vfp:(frame ref), caller:string, vars:((string * int ref) list)};

	(* DATATYPE tree : Type for representing the nesting heirarchy of the
	* procedures defined in the program.
	*)
	datatype 'a tree = Node of 'a * ('a tree list);
	exception unknown;

	val stack = ref ([]: frame list);

	fun search [] s = false
	|search (x::xs) s = (x=s) orelse (search xs s);
	
	val break = fn (F f) => f | NULL => raise unknown;
	val chomp = fn s => (implode(rev(tl(rev(explode s)))));

	fun buildTree (P p) = Node(#aname(p),map buildTree (#eproclist(p)));
	fun t2l (Node(t,tl)) = t::(foldl List.@ [] (map t2l tl));
	val chop = fn (Node(t,_)) => t;

	fun callable s l (Node(t,[])) = if(s=t) then (s::l) else []
	|callable s l (Node(t,tl)) = if(s=t) then ((map chop tl)@l)
					else 
					let
						val l0 = (map chop tl)@l
						val l1 = (map (callable s l0) tl)
						val l2 = foldr List.@ [] l1
					in l2
					end;
	
	fun findFather s f (Node(t,[])) = if(s=t) then f else ""
	|findFather s f (Node(t,tl)) = if(s=t) then f
					else
					let
						val l0 = map (findFather s t) tl
						val l1 = foldr String.^ "" l0
					in l1
					end;

	fun findNode s (p as (P pr)) = if(s = #aname(pr)) then p
				else
					let
						fun help [] = raise unknown
						|help (x::xs) = (findNode s x) handle unknown => (help xs)
						
						val p1 = help (#eproclist(pr))
					in p1
					end;

	local
	fun lookup [] s = (print("Varaible "^s^" not found.\n"); raise unknown)
	|lookup ((x,y,n)::xs) s = if(x=s) then !n else (lookup xs s)
	in
	val evalexp = fn l => let
				val _ = print "Enter the expression to be assigned: "
				val e = Option.valOf(inputLine stdIn)
				val file = openOut "temp$$"
				val _ = output(file,e)
				val _ = closeOut file
				val pt = compile "temp$$"
				val n = calculate (lookup l) pt
			in n end
			handle unknown => raise unknown
				| BuildAstError => (print "Malformed expression!\n";raise unknown)
	end;

	local
	fun filter s l [] = l
	|filter s l ((x,y,n)::xs) = if(x=s) then (filter s l xs) else (filter s ((x,y,n)::l) xs)
	in
	fun values NULL = []
	|values (F f) = let
				val p = #name(f)
				val l1 = map (fn (x,n) => (x,p,n)) (#vars(f))
				val l2 = values (!(#vfp(f)))
				val l3 = foldr (fn ((x,y,n),l) => (filter x [] l)) l2 l1
			in l1@l3
			end
	end;
	
	local
	fun getParam l [] = l
	|getParam l (x::xs) = let
					val _ = print ("For parameter "^x^", ")
					val n = evalexp (values (hd(!stack)))
				in
					getParam ((x,ref n)::l) xs
				end
				handle unknown => (print "Re-Enter--\n";(getParam l (x::xs)))
	val genvarlist= fn (l1,l2) => ((map (fn x => (x,ref 0)) l2)@(getParam [] l1))
	in
	val buildFrame = fn (P p) => fn fp => fn s => F{name = #aname(p), 
							vfp = fp,
							caller = s,
							vars = genvarlist(#bparam(p),#cvarlist(p))
							}
	end;

	val line = "_________________\n";
	val showFrame = fn NULL => () 
			| (F f) => let
				val p = #name(f)
				fun showVs [] = print (line)
				|showVs ((x,n)::xs) = (
							print(line^"[Var: "^x^" = "^(Int.toString (!n))^" ]\n");
							(showVs xs)
							)
				val vs = #vars(f)
				val f1 = !(#vfp(f))
				val sc = #caller(f)
				val parent = if (f1 = NULL) then "System" else (#name(break f1))
				in
				(
					print ("Stack Frame for Procedure \""^p^"\"\n");
					print (line^"[fp - Dynamic Link : "^sc^" ]\n");
					showVs vs;
					print ("[fp - Static Link : "^parent^" ]\n"^line)
				)
				end;

	fun printer [] = print "\n"
	|printer ((x,y,n)::xs) = (
					print("\n"^("Name: "^x^"; Owner: "^y^"; Value: "^(Int.toString (!n))^".\n"));
					(printer xs)
				);
	
	local
	fun patch _ [] = false
	|patch (x,n) ((y,p)::ys) = 
	if(x=y) then 
	(
		print ("Updating "^x^" from "^(Int.toString(!p))^" to "^(Int.toString(n))^".\n");
		p:=n;
		true
	)
	else patch (x,n) ys
	
	fun update [] l m = m
	|update ((x,n)::xs) l1 m = if(patch (x,n) l1) then (update xs l1 m) else (update xs l1 ((x,n)::m))
	in
	fun execute NULL [] = ()
	|execute NULL li = ( print "The following variables were not found:\n";
				print (foldr (fn ((x,n),l) => x^" "^l) ".\n" li)
			)
	|execute (F f) li = execute (!(#vfp(f))) (update li (#vars(f)) [])
	end;

	fun trace s r NULL = raise unknown
	|trace s r (F f) = if(s = #name(f)) then r else let val p = #vfp(f) in (trace s p (!p)) end;

	val dummy = P {aname="",bparam=[],cvarlist=[],dasmtlist=[],eproclist=[]};
	val program = ref dummy;
	val ptree = ref (Node("",[]));
	val dfp = ref "";
	val loaded = ref false;

	val load = fn p => (
				program := p;
				ptree := buildTree p;
				stack := [];
				dfp := "";
				loaded := true;
				print ("Program Loaded.\n")
			);
	val reset = fn () => (
				program := dummy;
				stack := [];
				dfp := "";
				loaded := false;
				print ("Program Unloaded.\n")
			);        

	val listproc = fn l => (print "Callable procedures are (in nesting order)\n";
				print (foldr (fn (x,s) => x^" "^s) ".\n" l)
			);

	fun main () = (
			print "\n\t...Main Menu...\n\n";
			print "1. List the Procedures that can be called.\n";
			print "2. Call a Procedure.\n";
			print "3. Return from Current Procedure.\n";
			print "4. View all accessible variables.\n";
			print "5. Update a variable.\n";
			print "6. Display stack-frame of Current Procedure.\n\n";
			print "Enter your choice : ";
			let
			val s = Option.valOf(inputLine stdIn)
			val n = Option.valOf(Int.fromString(s))
			in
			case n of 
			1 => (listproc (callable (!dfp) [] (!ptree)); main())
			| 2 => let
					val l = (callable (!dfp) [] (!ptree))
					val _ = listproc l
					val _ = print "Name of the procedure: "
					val s1 = chomp(Option.valOf(inputLine stdIn))
				in
				if (search l s1) then 
					let 
					val p = findNode s1 (!program)
					val s2 = findFather s1 "" (!ptree)
					val fp = hd(!stack)
					val fp2 = buildFrame p (trace s2 (ref fp) fp) (!dfp)
					in
					(
						dfp:=s1;
						stack := (fp2::(!stack));
						print("Executing body of procedure "^s1^":\n");
						execute fp2 (#dasmtlist(crack(p)))
					)end
				else print "Undefined or Inaccessible procedure!\n";
				main()
				end
			| 3 => if (length(!stack))=1 then reset()
				else
				(
					stack := tl(!stack);
					print ("Returning from "^(!dfp)^" to ");
					dfp := #name(break (hd (!stack)));
					print ((!dfp)^".\n");
					main()
				)
			| 4 => (
					print ("Variables accessible from "^(!dfp)^":\n");
					printer (values (hd(!stack)));
					main()
				)
			| 5 => (
					print ("Enter the name of the variable: ");
					let
						val s = chomp(Option.valOf(inputLine stdIn))
						val f = hd(!stack) 
					in
					execute f [(s,evalexp (values f))]
					end
					handle x => print "Updation Aborted.\n";
					main()
				)
			| 6 => (showFrame (hd(!stack)); main())
			| n => ((print "Invalid choice!\n");main())
			end
			handle Option.Option => (print ("Invalid choice!\n");(main()))
		)

	fun start () = if(not(!loaded)) then (print "Program not loaded. Please load a program for execution\n.")
		else
			(
				print "ImpSim Version 0.0\n\n";
				print "1. Start Execution of the main program (Top most procedure)\n";
				print "2. Exit\n\n";
				print "Enter your choice : ";
				let
				val s = Option.valOf(inputLine stdIn)
				val n = Option.valOf(Int.fromString(s))
				in
				if(n=1) then
				(
					stack := [buildFrame (!program) (ref NULL) "System"];
					dfp := #aname(crack(!program));
					print("Executing body of procedure "^(!dfp)^":\n");
					execute (hd(!stack)) (#dasmtlist(crack(!program)));
					main()
				)
				else reset()
				end
				handle Option.Option => (print ("Invalid choice!");(start()))
			);
end;

(* EOF *)
